--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Datapath structural description
--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_signed.all; -- needed for comparison instructions SLTxx
use IEEE.Std_Logic_arith.all; -- needed for comparison instructions SLTxx
use work.p_MR4.all;

entity datapath is
	port
	(
		ck, rst: in std_logic;
		i_address: out reg32;
		instruction: in reg32;
		d_address: out reg32;
		data_in: in reg32;
		data_out: out reg32;
		uins: in microinstruction;
		end_mul, end_div: out std_logic;
		IR_OUT: out reg32
	);
end datapath;

architecture datapath of datapath is

	signal	incpc, pc, npc, IR,  result, R1, R2, RS, RT, RIN, ext16, cte_im, IMED, op1, op2, outalu, RALU, P_Hi,
			P_Lo, R_Hi, Q_Lo, MDR, mdr_int, dtpc, D_Hi, D_Lo,  Hi, Lo: reg32;
	signal adD, adS: std_logic_vector(4 downto 0);
	signal inst_branch, inst_grupo1, inst_grupoI, Hi_Lo_ce,  end_mul_int, end_div_int: std_logic;
	signal salta: std_logic;

begin

-- auxiliary signals
	inst_branch  <= '1' when uins.i=BEQ or uins.i=BGEZ or uins.i=BLEZ or uins.i=BLTZ or uins.i=BNE else '0';
	inst_grupo1  <= '1' when uins.i=ADDU or uins.i=SUBU or uins.i=AAND or uins.i=OOR or uins.i=XXOR or uins.i=NNOR else '0';
	inst_grupoI  <= '1' when uins.i=ADDIU or uins.i=ANDI or uins.i=ORI or uins.i=XORI else '0';

--==============================================================================
-- first_stage
--==============================================================================
	incpc <= pc + 4;
	RNPC: entity work.regnbit port map(ck => ck, rst => rst, ce => uins.CY1, D => incpc, Q => npc);
	RIR: entity work.regnbit  port map(ck => ck, rst => rst, ce => uins.CY1, D => instruction, Q => IR);
	IR_OUT <= ir ;    -- IR is the datapath output signal to carry the instruction
	i_address <= pc;  -- connects PC output to the instruction memory address bus

--==============================================================================
-- second stage
--==============================================================================

	-- signal to be written into the register bank
	RIN <= npc when (uins.i=JALR or uins.i=JAL) else result;

	-- register bank write address selection
	adD <=	"11111" when uins.i=JAL else -- JAL writes in register $31
			IR(15 downto 11) when inst_grupo1='1' or uins.i=SLTU or uins.i=SLT or uins.i=JALR or uins.i=MFHI or uins.i=MFLO else
			IR(20 downto 16);       -- inst_grupoI='1' or uins.i=SLTIU or uins.i=SLTI or uins.i=LW or  uins.i=LBU  or uins.i=LUI, or default

	-- The then clause is only used for logic shifts with shamt field
	adS <= IR(20 downto 16) when uins.i=SSLL or uins.i=SSRA or uins.i=SSRL else IR(25 downto 21);

	REGS: entity work.reg_bank port map
			(ck => ck, rst => rst, wreg => uins.wreg,  AdRs => adS, AdRt => ir(20 downto 16), adRD => adD,  RD => RIN, R1 => R1, R2 => R2);

	-- sign extension
	ext16 <=  x"FFFF" & IR(15 downto 0) when IR(15)='1' else x"0000" & IR(15 downto 0);

	-- Immediate constant
	cte_im <=	ext16(29 downto 0)  & "00"     when inst_branch='1'     else                    -- branch address adjustment for word frontier
				"0000" & IR(25 downto 0) & "00" when uins.i=J or uins.i=JAL else                -- J/JAL are word addressed. MSB four bits are defined at the ALU, not here!
				x"0000" & IR(15 downto 0) when uins.i=ANDI or uins.i=ORI  or uins.i=XORI else   -- logic instructions with immediate operand are zero extended
				ext16;                                                                          -- The default case is used by addiu, lbu, lw, sbu and sw instructions

	-- second stage registers
	REG_S: entity work.regnbit   port map(ck => ck, rst => rst, ce => uins.CY2, D => R1,     Q => RS);
	REG_T: entity work.regnbit   port map(ck => ck, rst => rst, ce => uins.CY2, D => R2,     Q => RT);
	REG_IM: entity work.regnbit  port map(ck => ck, rst => rst, ce => uins.CY2, D => cte_im, Q => IMED);


--==============================================================================
-- third stage
--==============================================================================
	-- select the first ALU operand
	-- op1 <= pc  when inst_branch='1' else  -- use this below for code generated with the PCSPIM simulator
	-- op1 <= npc  when inst_branch='1' else  -- use this below for code generated with the MARS simulator
	op1 <= npc  when inst_branch='1' else RS;

	-- select the second ALU operand
	-- RT when grupo1 instructions or SLT/SLTU or JR or shifts with no shamt field
	op2 <= RT when inst_grupo1='1' or uins.i=SLTU or uins.i=SLT or uins.i=JR  or uins.i=SLLV or uins.i=SRAV or uins.i=SRLV else IMED;

	-- ALU instantiation
	inst_alu: entity work.alu port map (op1 => op1, op2 => op2, outalu => outalu, op_alu => uins.i);

	-- ALU register
	REG_alu: entity work.regnbit   port map(ck => ck, rst => rst, ce => uins.walu, D => outalu, Q => RALU);

	end_mul <= end_mul_int;

-- Multiplier instantiation
	inst_mult: entity work.mult_serial port map(clock=>ck, start=>uins.init_mul, Rs=>RS, B_Rt=>RT, end_mul=>end_mul_int, P_Hi=>P_Hi, A_Lo=>P_Lo);

	end_div <= end_div_int;

-- Divider instantiation
	inst_div: entity work.div_serial port map
	(
		reset=>rst, clock=>ck, start=>uins.init_div, Rs=>RS, B_Rt=>RT, end_div=>end_div_int, P_Hi=>R_Hi, A_Lo=>Q_Lo
	);

-- select source for the Hi Register
	D_Hi <= P_Hi  when uins.i=MULTU else  R_Hi;

-- select source for the Lo Register
	D_Lo <= P_Lo  when uins.i=MULTU else  Q_Lo;

-- Hi and Lo registers write enable
	Hi_Lo_ce <= '1' when (uins.walu='1' and ((uins.i=DIVU and end_div_int='1') or (uins.i=MULTU and end_mul_int='1'))) else '0';

-- Hi register
	REG_Hi: entity work.regnbit port map(ck=>ck, rst=>rst, ce=>Hi_Lo_ce, D=>D_Hi, Q=>Hi);

-- Lo register
	REG_Lo: entity work.regnbit port map(ck=>ck, rst=>rst, ce=>Hi_Lo_ce, D=>D_Lo, Q=>Lo);

-- evaluation of conditions to take the branch instructions
	salta <=  '1' when ( (RS=RT and uins.i=BEQ) or (RS>=0 and uins.i=BGEZ) or (RS<=0 and uins.i=BLEZ) or (RS<0 and uins.i=BLTZ) or (RS/=RT and uins.i=BNE) ) else '0';

	result <=	Hi when uins.i=MFHI else
				Lo when uins.i=MFLO else
				MDR when uins.i=LW  or uins.i=LBU else
				RALU;

--==============================================================================
-- fourth stage
--==============================================================================

	d_address <= RALU;
	data_out <= RT;

-- single byte reading from memory
	mdr_int <= data_in when uins.i=LW  else x"000000" & data_in(7 downto 0);

	RMDR: entity work.regnbit  port map(ck => ck, rst => rst, ce => uins.wmdr, D => mdr_int, Q => MDR);

--==============================================================================
-- fifth stage
--==============================================================================
	dtpc <= result when(inst_branch='1' and salta='1') or uins.i=J or uins.i=JAL or uins.i=JALR or uins.i=JR else npc;

	-- Data memory starting address: beware of the OFFSET!
	-- rpc: entity work.regnbit generic map(INIT_VALUE => x"00400020")   use this below for code generated with the PCSPIM simulator
	-- rpc: entity work.regnbit generic map(INIT_VALUE => x"00400000")   use this below for code generated with the MARS simulator
	rpc: entity work.regnbit generic map(INIT_VALUE => x"00400000")
		port map(ck => ck, rst => rst, ce => uins.wpc, D => dtpc, Q => pc);

end datapath;
