--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- Multiplier - This is an implementation of a 32-bit serial multiplier
--              as proposed in Hennessy & Patterson textbook
--              - The implementation waits for a 1 clock cycle initialization
--              input control signal, then it computes the multiplication in
--              the next 65 clock cycles, in the last of these generating a
--              1 clock cycle output signal (end_mul) to indicate
--              the end of the multiplication, which result is then
--              available at the data outputs.
--++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

entity mult_serial is
	port
	(
		clock : in STD_LOGIC;
		start : in STD_LOGIC;                          -- typically activated for exactly 1 clock cycle
		Rs : in STD_LOGIC_VECTOR(31 downto 0);         -- multiplicand
		B_Rt : in STD_LOGIC_VECTOR(31 downto 0);       -- multiplier
		end_mul : out STD_LOGIC;                       -- end of multiplication, active for exactly 1 clock cycle
		P_Hi : out STD_LOGIC_VECTOR(31 downto 0);      -- higher order word of result
		A_Lo : out STD_LOGIC_VECTOR(31 downto 0)       -- lower order word of result
	);
end mult_serial;

architecture mult_serial of mult_serial is

	type states is (Init, Sum, Displace);
	signal PS, NS : states;
	signal B, P_Hi_int, parcel : std_logic_vector(32 downto 0);
	signal A_Lo_int : std_logic_vector(31 downto 0);
	signal cont : std_logic_vector(4 downto 0);

begin

	P_Hi <= P_Hi_int(31 downto 0);
	A_Lo <= A_Lo_int;

	-- multiplication control state machine
	process(clock)
	begin
		if clock'event and clock='1' then
			PS <= NS;
		end if;
	end process;

	process(PS, start, cont)
	begin
		case PS is
			when Init =>
				if start='1' then
					NS <= Sum;
				else
					NS <= Init;
				end if;

			when Sum =>
				NS <= Displace;

			when Displace =>
				if(cont = "00000")
					then NS <= Init;
				else
					NS <= Sum;
				end if;
		end case;
	end process;

	-- combinational element of the datapath - conditional adder
	parcel <= P_Hi_int + ('0' & B_Rt) when A_Lo_int(0)='1' else P_Hi_int;

	-- sequential elements of the datapath
	-- P_Hi, A_Lo registers, cont counter, end_mul output
	process(clock)
	begin
		if clock'event and clock = '1' then
			case PS is
				when Init =>
					if start = '1' then
						P_Hi_int <= (others=>'0');
						A_Lo_int <= Rs;
						cont <= (others=>'0');
					end if;
					end_mul <='0';

				when Sum =>
					cont <= cont-1;
					P_Hi_int <= parcel;

				when Displace =>
					if cont = "00000" then
						end_mul <='1';
					end if;
					P_Hi_int <= '0' & P_Hi_int (32 downto 1);
					A_Lo_int <= P_Hi_int (0)& A_Lo_int (31 downto 1);
			end case;
		end if;
	end process;

end mult_serial;

